#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
############################################################################
#
# (c) 2006 Florian Birée aka Thesa <florian.biree@laposte.net>
# (c) 2006 Achraf Cherti aka Asher256 <achrafcherti@gmail.com>
#
#  Website: <URL:http://casetta.tuxfamily.org>
#
# Version 0.2.0
#
# changelog:
#
# # 0.2.0 version:
#
#   This file was split in two file: this file is the CORE part, and the 
#   casetta_cli.py is the CLI part. See this file for changelog of this part.
#
#   o Change the formats and datas way of internal description
#   o Add a check_name function
#   o Add a copy function to copy a Data
#   o Add a split_by_data_type function
#   o Add a date metadata (datetime.date type)
#   o All save functions can now save all or a part of datas (export flag)
#   o Change the way to work of casio_import and casio_export
#   o New exception system
#   o casio_import fill the date metadata
#   o New data type: backups
#   o cafix format can managed backup datas
#   o casio format can managed backup datas
#   o Add the fxb format
#   o Add the fxm format (read only)
#   o Add the date metadata management for newcat format (newcat 0.2)
#   o newcat format can managed backup (in input/output) (newcat 0.2)
#   o Add a password metadata (string type), managed in cat and newcat 0.2
#   o Add the fxd format (read only)
#   o Add the g1r format (read only - with base management)
#   o Correct a bug in text/cat encode/decode about *, /, "\DispR-Tbl" and
#     "\DrawWeb"
#   o Allow a file format to have many extensions
#   o Internal base programs management
#   o Base programs management in followed format:
#      newcat, cat, fxd, fxi
#   o Management of the Calculator Text Format for programs, in read and write
#     access, following the pre-1 (Revision A) version, avaible in:
#     <URL:http://members.lycos.co.uk/rfam/archive/ctf/ctf.html>
#   o Auto-detection of base programs in cafix and ctf formats
#   o Add a title option in text/cat encode/decode function to convert equivs 
#     in titles and passwords
#   o Management of specials characters in title and passwords of cat and 
#     newcat formats
#   o New way to parse/write cat and newcat formats, many bug fix (such as the
#     capacity bug)
#   o New way to manage transfers tool
#   o New transfer tool: CaS
#   o New transfer tool: other
#   o Replace os.tmpnam() by os.tempnam()
#   o Disable the CaS transfer tool (the other transfer tool can be used 
#     with CaS)
#   o Use tempfile.mk[ds]temp() instead of os.tempnam() (security reason)
#
# # 0.1.2 version:
#
#   o 'or' bug fix in the fxiDecode function
#   o change the way to found the name of a data in the fxi 
#     format (work with any name length)
# 
# # 0.1.1 version:
#
#   o many bug fix
#
# # 0.1.0 version:
#
#   o first release
#
############################################################################
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# http://www.gnu.org/copyleft/gpl.html
#
############################################################################
"""Python module for conversions into casio programs files"""

__author__ = "Florian Birée, Achraf Cherti"
__version__ = "0.2.0"
__copyright__ = "Copyright (c) 2006, Florian Birée, Achraf Cherti"
__license__ = "GPL"
__revision__ = "$Revision: $"
# $Source: $
__date__ = "$Date: $"

# Imports
import os, sys
import re
import datetime
import base64
import tempfile
import catdic 
import ctfdic

# Other transfert tool variables
_other_tool_format = None
_other_tool_receive = None
_other_tool_send = None

# Build the raw2fxi dic
# Thanks to http://www.casioland.net/tutoriaux.php?idTuto=3&noHeader=1
r2f = {}
dep, chrNumber, fxi_index, fxi_index2 = 23, 255, 0, 0
chr_value = dep
while chrNumber >= 0:
    chr_value = chr_value + 1
    r2f[chrNumber] = chr_value
    chrNumber -= 1
    fxi_index += 1
    if fxi_index == 8:
        dep -= 8
        fxi_index2 += 1
        fxi_index = 0
        chr_value = dep
    if fxi_index2 == 4:
        fxi_index2 = 0
        dep += 8 * 8
        chr_value = dep

# Build the fxi2raw dic
f2r = {}
for ascii, equiv in r2f.items():
    f2r[equiv] = ascii

# Exceptions
class FormatCantOpen(Exception):
    """Exception raised when the open function is not found for format"""

    def __init__(self, format):
        Exception.__init__(self)
        self.format = format

    def __str__(self):
        return repr(self.format)

class ToolNotFound(Exception):
    """Exception raised when a tool cannot be used 
    
    (the open function is not found for this tool or the command is not 
    in the path)"""

    def __init__(self, tool):
        Exception.__init__(self)
        self.tool = tool

    def __str__(self):
        return repr(self.tool)

class FormatNotFound(Exception):
    """Exception raised when a function cannot select the right format"""
    pass

class DataNoManaged(Exception):
    """Exception raised when a format cannot manage some data type"""

    def __init__(self, data_type_list):
        Exception.__init__(self)
        self.data_type_list = data_type_list

    def __str__(self):
        return repr(self.data_type_list)

class NoCommand(Exception):
    """Exception raised when a command is not found in the PATH"""

    def __init__(self, command):
        Exception.__init__(self)
        self.command = command

    def __str__(self):
        return repr(self.command)

class ToolNotConfigured(Exception):
    """Exception raised when a mistake arrive using a transfer tool"""
    pass

# CORE part

class Data:
    """Data manipulation class"""

    def __init__(self):
        """Make an empty data"""
        self.name = ''
        self.raw_data = ''
        self.dType = 'none'
        self.date = None
        self.password = ""
        self.use_base = False

    def __repr__(self):
        """Descript this data"""
        return "<Data '%s', type: '%s'>" % (self.name, self.dType)

    def get_text(self):
        """Convert raw_data in cat-like text (only for programs)"""
        if self.dType == 'program':
            return text_encode(self.raw_data)

    def set_text(self, text):
        """Convert cat-like text in raw_data (only for programs)"""
        if self.dType == 'program':
            self.raw_data = text_decode(text)

class FileData:
    """Data file manipulation class"""

    def __init__(self):
        """Make an empty FileData"""
        self.data = []
        self.export = []

    def __repr__(self):
        """Descript this data"""
        out = "<File of %s casio data:\n" % str(len(self.data))
        for data_index in range(len(self.data)) :
            out += "#%s: %s\n" % (str(data_index), \
                                  self.data[data_index].__repr__())
        out += '>'
        return out

    def __getitem__(self, index):
        """the same of self.data[index]"""
        return self.data[index]

    def del_data(self, delete_index):
        """Del the data with the index i"""
        del(self.data[delete_index])
        # clear the export list because index list has changed
        self.export = []

    def new_data(self):
        """Make a new empty data, and return it index"""
        self.data.append(Data())
        return len(self.data) - 1

    def save(self, filename, format='unknown', ignore_warnings=False, \
             export=False):
        """Save in (a) file(s) the data.

        If all data aren't managed by this format, raise DataNoManaged
        with the list of data type not managed (['all'] for all ;-) )
        Use ignore_warnings = True to save already in this format.
        
        """
        if format == '' or format == 'unknown':
            format = choose_format(filename)
            if format == 'unknown':
                raise FormatNotFound
        if not ignore_warnings :
            no_managed_data = check_no_managed_data(self, format, export)
            if no_managed_data != []:
                raise DataNoManaged(no_managed_data)
        saveFunc[format](filename, self, export)

    def add_data(self, file_data):
        """Add data from another file_data

        Return the list of new index
        (You should check the name of those datas)
        
        """
        length = len(self.data)
        self.data += file_data.data
        return range(length, len(self.data))

    def add(self, filename, format = 'unknown'):
        """Open datas from filename and add them in the file .

        Return the list of new index
        (You should check the name of those datas)
        
        """
        if format == 'unknown' :
            format = choose_format(filename)
            if format == 'unknown':
                raise FormatNotFound
        return self.add_data(openFunc[format](filename))

    def list_data_types(self, export=False):
        """Return the list of data types used"""
        ret = []
        if export:
            indices = self.export
        else :
            indices = range(len(self.data))
        for index in indices :
            if not self[index].dType in ret :
                ret.append(self[index].dType)
        return ret

    def check_name(self, data_index):
        """Return True if the name of this data is correct"""
        # check if already employed
        for index in range(len(self.data)) :
            if index != data_index and \
               self.data[index].dType == self.data[data_index].dType and \
               self.data[index].name == self.data[data_index].name :
                return False
        # check the syntax of the name
        if re.search(dataName.get(self.data[data_index].dType), \
                     self.data[data_index].name) == None :
            return False
        return True

    def copy_data(self, data_index):
        """Copy data from dataIndex.

        Return the new index
        (You should check the name of the copy)
        
        """
        new_index = self.new_data()
        self.data[new_index].name = self.data[data_index].name
        self.data[new_index].raw_data = self.data[data_index].raw_data
        self.data[new_index].dType = self.data[data_index].dType
        self.data[new_index].date = self.data[data_index].date
        self.data[new_index].password = self.data[data_index].password
        return new_index

    def split_by_data_type(self):
        """Split this file_data by data type.

        Return a dictionary of file data
        
        """
        ret = {}
        for data in self.data:
            if not data.dType in ret.keys():
                ret[data.dType] = FileData()
            new_index = ret[data.dType].new_data()
            ret[data.dType].data[new_index].name = data.name
            ret[data.dType].data[new_index].raw_data = data.raw_data
            ret[data.dType].data[new_index].dType = data.dType
            ret[data.dType].data[new_index].date = data.date
            ret[data.dType].data[new_index].password = data.password
        return ret

# Decode/Encode functions
def fxi_encode(raw_data):
    """Convert raw data to fxi-like data"""
    out = ''
    for car in range(len(raw_data)):
        out += chr(r2f[ord(raw_data[car])])
    return out

def fxi_decode(fxi_data):
    """Convert fxi-like data to raw data"""
    out = ''
    for car in range(len(fxi_data)):
        out += chr(f2r[ord(fxi_data[car])])
    # Fix the 'or' bug (some 'or' strings was replaced 
    # by a 0xAA string ##v0.1.2
    out = out.replace(chr(0xAA), 'or')
    return out

def cat_encode(raw_data, title=False):
    """Convert raw data to cat-like data"""
    if title:
        out = raw_data
        for raw_car, cat_equiv in catdic.title_equiv.items():
            out = out.replace(raw_car, cat_equiv)
        return out
    out = raw_data
    # convert * /
    out = out.replace("*", "\\aster")
    out = out.replace("\xf7\\aster", "\xf7*")
    out = out.replace("/", "\\slash")
    out = out.replace("\xf7\\slash", "\xf7/")
    #Convert ctokens in tokens
    ctoken_list, token_for = catdic.get_r2c_data()
    for ctoken in ctoken_list:
        out = out.replace(ctoken, token_for[ctoken])
    # Remove the end caract (chr(255))
    out = out.replace(chr(255), '')
    # Change end line (chr(13) to chr(13)+chr(10)
    out = out.replace(chr(13), chr(13) + chr(10))
    
    return out

def cat_decode(cat_data, title=False):
    """Convert cat-like data to raw data"""
    if title:
        out = cat_data
        for raw_car, cat_equiv in catdic.title_equiv.items():
            out = out.replace(cat_equiv, raw_car)
        return out
    out = cat_data
    # Change end line chr(13)+chr(10) in chr(13)
    out = out.replace(chr(13) + chr(10), chr(13))
    # Save "\\DispR-Tbl" and "\\DrawWeb"
    out = out.replace("\\DispR-Tbl", "\\save_drtbl")
    out = out.replace("\\DrawWeb", "\\save_DrawWeb")
    #Convert ctokens in tokens
    token_list, ctoken_for = catdic.get_c2r_data()
    for token in token_list:
        out = out.replace(token, ctoken_for[token])
    # convert * /and save "\\DispR-Tbl" and "\\DrawWeb"
    out = out.replace("\\aster", "*")
    out = out.replace("\\slash", "/")
    out = out.replace("\\save_drtbl", chr(247)+chr(47))
    out = out.replace("\\save_DrawWeb", chr(247)+chr(42))
    # Add the last caract (chr(255))
    out += chr(255)
    return out

def text_encode(raw_data, title=False):
    """Convert raw data to Editable text (NewCat format)"""
    if title:
        out = raw_data
        for raw_car, cat_equiv in catdic.title_equiv.items():
            out = out.replace(raw_car, cat_equiv)
        return out
    out = raw_data
    # convert * /
    out = out.replace("*", "\\aster")
    out = out.replace("\xf7\\aster", "\xf7*")
    out = out.replace("/", "\\slash")
    out = out.replace("\xf7\\slash", "\xf7/")
    #Convert ctokens in tokens
    ctoken_list, token_for = catdic.get_r2c_data(True)
    for ctoken in ctoken_list:
        out = out.replace(ctoken, token_for[ctoken])
    # Remove the end caract (chr(255))
    out = out.replace(chr(255), '')
    # Change end line chr(13) in chr(10)
    out = out.replace(chr(13), '\n')
    return out

def text_decode(text_data, title=False):
    """Convert Editable text (NewCat format) to raw data"""
    if title:
        out = text_data
        for raw_car, cat_equiv in catdic.title_equiv.items():
            out = out.replace(cat_equiv, raw_car)
        return out
    out = text_data
    # Change end line chr(10) in chr(13)
    out = out.replace('\n', chr(13))
    # Save "\\DispR-Tbl" and "\\DrawWeb"
    out = out.replace("\\DispR-Tbl", "\\save_drtbl")
    out = out.replace("\\DrawWeb", "\\save_DrawWeb")
    #Convert ctokens in tokens
    token_list, ctoken_for = catdic.get_c2r_data(True)
    for token in token_list:
        out = out.replace(token, ctoken_for[token])
    # convert * / and save "\\DispR-Tbl" and "\\DrawWeb"
    out = out.replace("\\aster", "*")
    out = out.replace("\\slash", "/")
    out = out.replace("\\save_drtbl", chr(247)+chr(47))
    out = out.replace("\\save_DrawWeb", chr(247)+chr(42))
    # Add the last caract (chr(255))
    out += chr(255)
    return out

# File format manager

def check_no_managed_data(file_data, format, export=False) :
    """ Return the list of data type in file_data no managed in format.

    Return [] if all data type are managed, ['all'] if nothing is managed
    
    """
    ret = []
    for data in file_data.list_data_types(export) :
        if not data in dataManaged.get(format) :
            ret.append(data)
    ret.sort()
    fd_list = file_data.list_data_types()
    fd_list.sort()
    if fd_list == ret :
        ret = ['all']
    return ret
    
def choose_format(filename):
    """Return the format of filename (with extension)"""
    for format, list_ext in formatExt.items() :
        for ext in list_ext:
            if filename[-len(ext)-1:].lower() == '.' + ext :
                return format
    return 'unknown'

def open_file(filename, format=''):
    """Open a file using the function corresponding to the format.

    Raise an FormatCantOpen error if the openFunc has not format key
    Return None if the file cannot be open
    
    """
    if format == '' :
        format = choose_format(filename)
    try :
        return openFunc[format](filename)
    except KeyError:
        raise FormatCantOpen(format)

def fxi_open(filename):
    """Open a file in the fxi Format (new function using re)"""
    fxi_file = open(filename, 'r')
    # Make a new FileData
    file_data = FileData()
    # Get the raw data
    ctn = fxi_decode(fxi_file.read())
    fxi_file.close()
    # Make the regular expression
    data_pattern = re.compile('''
    (TXT)           # data format (but fxi txt is not really text :-/)
    .               # separator
    (PG)            # type of data
    .{4}            # 4 characters
    ([^\xff]{1,8})  # data name
    \xff{16,23}     # between 16 and 23 (16+8-len(data name)) 0xFF
    (NL|BN)         # Option1 (Normal or Base)
    \xff{12}        # 12 0xFF
    ([^\xff]+\xff)  # any non 0xFF characters ended by one 0xFF : raw data
    ''', re.VERBOSE)
    # Extract records from raw data using the regular expression
    data_list = data_pattern.findall(ctn)
    for record in data_list:
        # record = ('FORMAT', 'DATA TYPE', 'FILE NAME', 'OPTION', 'RAW DATA')
        new_index = file_data.new_data()
        if record[1] == 'PG' :
            file_data[new_index].dType = 'program'
            file_data[new_index].name = record[2]
            if record[3] == 'BN':
                file_data[new_index].use_base = True
            else:
                file_data[new_index].use_base = False
            file_data[new_index].raw_data = record[4]
    return file_data

def fxb_open(filename):
    """Open a backup in FX-Interface backup (fxb) format"""
    fxb_file = open(filename, 'r')
    ctn = fxb_file.read()
    fxb_file.close()
    file_data = FileData()
    # Delete the 10 FF in the header
    ctn = ctn.replace(10*chr(0xFF), '', 1)
    # Split by 9 FF
    bak = (9*chr(0xFF)).join(ctn.split(9*chr(0xFF))[1:])
    new_index = file_data.new_data()
    file_data[new_index].raw_data = bak
    file_data[new_index].name = os.path.basename(filename)[:-4]
    file_data[new_index].dType = 'backup'
    return file_data

def fxb_save(directory, file_data, export=False):
    """Save a file in the fxb Format"""
    folder = os.path.dirname(directory)
    # For each data
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for data in [file_data[index] for index in index_list if \
                 file_data[index].dType == 'backup']:
        # make a filename from the metadata
        if len(file_data.data) == 1 and os.path.basename(directory) != '':
            # try to use directory as name
            filename = directory
        else:
            if data.date == None:
                date_prefix = ''
            else:
                date_prefix = data.date.strftime('%Y%m%d') + '-'
            filename = os.path.join(folder, date_prefix + data.name + '.fxb')
        # fxb header:
        ctn = base64.decodestring("""
iAABAAAARlggQkFDSyBVUCBGSUxFAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA+AAACwAXAIEaMgEAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAZAAAAP7///9cAAAAAAAAEUNBU0lPIEZYIEJBQ0sgVVAg
REFUQQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAuAQAAAAEABgAl
AAEAAAAAABAARGF0YQBA+AAAPPgAAAEAAAA0+AAAMPgAAE1FTQBCVQAA+AJCYWNrdXD/////////
////R1kzNTj//4AACCAAAP///////////w==""")
        ctn += data.raw_data
        cafix_file = open(filename, 'w')
        cafix_file.write(ctn)
        cafix_file.close()

def fxm_open(filename):
    """Open a backup in Fx-Interface Pro backup (fxm) format"""
    fxm_file = open(filename, 'r')
    ctn = fxm_file.read(67706)
    fxm_file.close()
    file_data = FileData()
    bak = ctn[4218:]
    new_index = file_data.new_data()
    file_data[new_index].raw_data = bak
    file_data[new_index].name = os.path.basename(filename)[:-4]
    file_data[new_index].dType = 'backup'
    return file_data

def fxd_open(filename):
    """Open a file in the fxd Format"""
    fxd_file = open(filename, 'r')
    # Build the FileData
    file_data = FileData()
    # Make the regular expression
    data_pattern = re.compile('''
    (PG)            # type of data
    .{4}            # 4 characters
    ([^\xff]{1,8})  # data name
    \xff{16,23}     # between 16 and 23 (16+8-len(data name)) 0xFF
    (NL|BN)         # Option1 (Normal or Base)
    \xff{13}        # 13 0xFF
    .{8}            # 8 characters
    ([^\xff]+\xff)  # any non 0xFF characters ended by one 0xFF : raw data
    ''', re.VERBOSE)
    # extract data from fxd_file
    data_list = data_pattern.findall(fxd_file.read())
    fxd_file.close()
    for record in data_list:
        # record = ('DATA TYPE', 'FILE NAME', 'OPTION', 'RAW DATA')
        new_index = file_data.new_data()
        if record[0] == 'PG' :
            file_data[new_index].dType = 'program'
            file_data[new_index].name = record[1]
            if record[2] == 'BN':
                file_data[new_index].use_base = True
            else:
                file_data[new_index].use_base = False
            file_data[new_index].raw_data = record[3]
    return file_data

def g1r_open(filename):
    """Open a file in the g1r Format"""
    g1r_file = open(filename, 'r')
    # Build the FileData
    file_data = FileData()
    # Make the regular expression
    data_pattern = re.compile('''
    (PROGRAM)      # type of data
    [\x00]{12}     # Separator
    [\x01]         # one 0x01 character
    system         # The 'system' string
    [\x00]{2}      # Separator
    ([^\x00]{1,8}) # data name
    [\x00]{0,7}    # to complete the data name
    [\x01]         # end of data name
    [\x00]{2}      # Separator
    .{2}           # two characters
    [\x00]{11}     # Separator
    (\x00|\x01)    # Base (01) or not (00)
    \x00           # Separator
    (.*[^\x00])    # any characters ended by a non 0x00 : raw data 
                   #(w/o the 0xFF end character)
    [\x00]{2}      # end of the record
    ''', re.VERBOSE)
    # extract data from g1r_file
    data_list = data_pattern.findall(g1r_file.read())
    g1r_file.close()
    for record in data_list:
        # record = ('DATA TYPE', 'FILE NAME', 'BASE', 'RAW DATA')
        new_index = file_data.new_data()
        if record[0] == 'PROGRAM' :
            file_data[new_index].dType = 'program'
            file_data[new_index].name = record[1]
            if record[2] == '\x01':
                file_data[new_index].use_base = True
            else:
                file_data[new_index].use_base = False
            file_data[new_index].raw_data = record[3] + '\x0d\xff'
    return file_data

def cat_open(filename):
    """Open a file in the cat Format"""
    # Build the FileData
    file_data = FileData()
    cat_file = open(filename, 'r')
    ctn = cat_file.read()
    cat_file.close()
    # Split data
    for record in ctn.split('%End' + chr(13)+chr(10)):
        if '%Data Record' in record:
            header = record.split('%Data Record' + chr(13)+chr(10))[0]
            data = record.split('%Data Record' + chr(13)+chr(10))[1]
            # for each data, set metadata
            data_type = 'unknown'
            format = 'raw'
            name = 'unknown'
            use_base = False
            for headline in header.split(chr(13)+chr(10)):
                label = headline.split(':')[0]
                try:
                    value = headline.split(':')[1]
                except IndexError:
                    value = ''
                if label == 'Data Type' and value == 'PG':
                    data_type = 'program'
                elif label == 'Format' and value == 'TXT':
                    format = 'txt'
                elif label == 'File Name':
                    name = value
                elif label == 'Password':
                    password = value
                elif label == 'Option1' and value == 'BN':
                    use_base = True
            # Creat the data
            new_index = file_data.new_data()
            file_data.data[new_index].name = cat_decode(name, True)
            file_data.data[new_index].dType = data_type
            file_data.data[new_index].password = cat_decode(password, True)
            file_data.data[new_index].use_base = use_base
            if format == 'raw':
                file_data.data[new_index].raw_data = data
            elif format == 'txt':
                file_data.data[new_index].raw_data = cat_decode(data)
    return file_data

def cat_save(filename, file_data, export=False):
    """Save a file in the cat Format"""
    cat_file = open(filename, 'w')
    ctn = ''
    # For each data, convert in the cat format
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for data in [file_data.data[index] for index in index_list]:
        data_type = ''
        if data.dType == 'program':
            data_type = "PG"
        if data_type != '':
            # Set the header and metadata
            ctn += "%Header Record"+chr(13)+chr(10)
            ctn += "Format:TXT"+chr(13)+chr(10)
            ctn += "Communication SW:0"+chr(13)+chr(10)
            ctn += "Data Type:" + data_type + chr(13)+chr(10)
            ctn += "Capacity:" + str(len(data.raw_data) + 2) + chr(13)+chr(10)
            ctn += "File Name:" + cat_encode(data.name, True) + chr(13)+chr(10)
            ctn += "Group Name:" + chr(13)+chr(10)
            ctn += "Password:" + cat_encode(data.password, True) + chr(13) + \
                                                                        chr(10)
            if data.use_base:
                ctn += "Option1:BN" + chr(13)+chr(10)
            else:
                ctn += "Option1:NL" + chr(13)+chr(10)
            ctn += "Option2:" + chr(13)+chr(10)
            ctn += "Option3:" + chr(13)+chr(10)
            ctn += "Option4:" + chr(13)+chr(10)
            ctn += "%Data Record" + chr(13)+chr(10)
            # Write cat data
            ctn += cat_encode(data.raw_data)
            ctn += chr(13)+chr(10) + "%End" + chr(13)+chr(10)
    # join in a cat file
    cat_file.write(ctn)
    cat_file.close()

def newcat_open(filename):
    """Open a file in the newcat Format"""
    # Build the file_data
    file_data = FileData()
    nc_file = open(filename, 'r')
    ctn = nc_file.read()
    nc_file.close()
    # Split data
    for record in ctn.split('%End\n'):
        if '%Data Record' in record:
            header, data = record.split('%Data Record\n')
            # for each data, set metadata
            data_type = 'unknown'
            format = 'RAW'
            name = 'unknown'
            date = None
            use_base = False
            for headline in header.split('\n'):
                label = headline.split(':')[0]
                try:
                    value = headline.split(':')[1]
                except IndexError:
                    value = ''
                if label == 'Format' :
                    format = value
                elif label == 'Data Type':
                    if value == 'PG':
                        data_type = 'program'
                    elif value == 'BK':
                        data_type = 'backup'
                elif label == 'File Name' or label == 'Data Name' or \
                     label == 'Variable Name':
                    name = value
                elif label == 'Date':
                    date = datetime.date(int(value[:4]), int(value[4:6]), \
                                         int(value[6:]))
                elif label == 'Password':
                    password = value
                elif label == 'Option1' and value == 'BN':
                    use_base = True
            # Creat the data
            new_index = file_data.new_data()
            file_data.data[new_index].name = text_decode(name, True)
            file_data.data[new_index].dType = data_type
            file_data.data[new_index].date = date
            file_data.data[new_index].password = text_decode(password, True)
            file_data.data[new_index].use_base = use_base
            if format == 'RAW':
                file_data.data[new_index].raw_data = data
            elif format == 'RAW64':
                file_data.data[new_index].raw_data = base64.decodestring(data)
            elif format == 'TXT':
                file_data.data[new_index].raw_data = text_decode(data)
    return file_data

def newcat_save(filename, file_data, export):
    """Save a file in the newcat Format"""
    nc_file = open(filename, 'w')
    ctn = ''
    # For each data, convert in the newcat format
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for data in [file_data.data[index] for index in index_list]:
        data_type = ''
        if data.dType == 'program':
            format = "TXT"
            data_type = "PG"
        elif data.dType == 'backup':
            format = "RAW64"
            data_type = "BK"
        if data_type != '':
            # Set the header and metadata
            ctn += "%Header Record\n"
            ctn += "Format:" + format + "\n"
            # ctn += "Communication SW:0\n"
            ctn += "Data Type:" + data_type + '\n'
            # ctn += "Capacity:" + str(len(data.raw_data)) + '\n'
            ctn += "File Name:" + text_encode(data.name, True) + '\n'
            # ctn += "Group Name:\n"
            ctn += "Password:" + text_encode(data.password, True) + "\n"
            if data.use_base:
                ctn += "Option1:BN\n"
            # ctn += "Option2:\n"
            # ctn += "Option3:\n"
            # ctn += "Option4:\n"
            if data.date != None :
                ctn += "Date:" + data.date.strftime("%Y%m%d") + "\n"
            ctn += "%Data Record\n"
            # Write *New* cat data
            if format == "TXT" :
                ctn += text_encode(data.raw_data)
            elif format == "RAW64" :
                ctn += base64.encodestring(data.raw_data)
            ctn += "%End\n"
    # join in a newcat file
    nc_file.write(ctn)
    nc_file.close()

def cafix_open(filename):
    """Open a file in the cafix Format"""
    # Build the FileData
    file_data = FileData()
    new_index = file_data.new_data()
    # Try to import metadata from the file name
    file_data.data[new_index].name = os.path.basename(filename)\
                       .replace('prog-', '').replace('.cafix', '')\
                       .replace('backup-', '')
    # Open raw data
    cafix_file = open(filename, 'r')
    file_data.data[new_index].raw_data = cafix_file.read()
    cafix_file.close()
    # Try to set the type:
    if os.path.basename(filename)[:5] == 'prog-':
        file_data.data[new_index].dType = 'program'
        file_data[new_index].use_base = is_base_program( \
                                                file_data[new_index].raw_data)
    elif os.path.basename(filename)[:7] == 'backup-':
        file_data.data[new_index].dType = 'backup'
    else:
        file_data.data[new_index].dType = 'unknown'
    return file_data

def cafix_save(directory, file_data, export=False):
    """Save a file in the cafix Format"""
    folder = os.path.dirname(directory)
    # For each data
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for data in [file_data.data[index] for index in index_list]:
        # make a filename from the metadata
        if len(file_data.data) == 1 and os.path.basename(directory) != '':
            # try to use directory as name
            filename = directory
        else:
            if data.dType == 'program':
                filename = os.path.join(folder, 'prog-' + data.name \
                                        + '.cafix')
            elif data.dType == 'backup':
                filename = os.path.join(folder, 'backup-' + data.name + \
                                        '.cafix')
            elif data.dType == 'unknown':
                filename = os.path.join(folder, 'unknown-' + data.name\
                           + '.cafix')
        # write raw data
        cafix_file = open(filename, 'w')
        cafix_file.write(data.raw_data)
        cafix_file.close()

def ctf_open(filename):
    """Open a file in the Calculator Text Format"""
    # Build the FileData
    file_data = FileData()
    # Open the file and get the content
    ctf_file = open(filename, 'r')
    ctn = ctf_file.read()
    ctf_file.close()
    # Uniformize end lien characters (and del wrong end line (\\\n)
    ctn = ctn.replace('\r\n','\n')
    ctn = ctn.replace('\r', '\n')
    ctn = ctn.replace('\\\n', '')
    # Begin to parse the file
    block_type = None # may be 'editor','oldprog','text','raw',None
    current_index = -1
    data = None
    for line in ctn.split('\n'):
        if line[:2] == '@@':
            # New block
            meta_data_pattern = re.compile(r'''
            ^           # Start of line
            @@          # Start of the block
            \s*         # CaS add a space (not in specification)
            ([\S]+)     # Block type
            \s+         # Separator
            ([^(]+)     # Any non ( character: name
            \s*         # the separator space
            (\(.*\))*   # options
            \s*         # some space
            $           # end of line
            ''', re.VERBOSE)
            meta_data_list = meta_data_pattern.findall(line)[0]
            block = meta_data_list[0].lower()
            name = meta_data_list[1]
            if name == '':
                name = 'oldprog'
            if name[-1] == ' ':
                name = name[:-1]
            options = meta_data_list[2].replace('(', '').replace(')', '')
            if block == 'prog' or block == 'program':
                if '"' in name:
                    block = 'editor'
                else:
                    block = 'oldprog'
            if block == 'oldprog':
                block_type = 'oldprog'
                current_index = file_data.new_data()
                file_data[current_index].dType = 'program'
                file_data[current_index].name = ctfdic.ctf_decode(name, True)
            elif block == 'editor':
                block_type = 'editor'
                current_index = file_data.new_data()
                file_data[current_index].dType = 'program'
                file_data[current_index].name = ctfdic.ctf_decode(name.\
                                                    replace('"', ''), True)
                if options != '':
                    file_data[current_index].password = options
            elif block == 'text':
                block_type = 'text'
                current_index = -1
            elif block == 'raw':
                block_type = 'raw'
                current_index = -1
            else:
                block_type = None
                current_index = -1
        elif line[:1] == '#':
            # Comment, ignored
            pass
        elif block_type == 'editor' or block_type == 'oldprog':
            # Delete inline comments
            re_com = re.compile('(\s+;[^"]*)$')
            line = re_com.sub('', line)
            # Replace space
            re_space = re.compile('(\s{2,})[^"]*$')
            line = re_space.sub(' ', line)
            re_space = re.compile('(\s+)$')
            line = re_space.sub('', line)
            # Parse tokens
            line = ctfdic.ctf_decode(line)
            # remove | (tokens separations)
            line = line.replace('|', '')
            # replace special characters @bnnn
            re_spe = re.compile('(@[bhxo].{3})')
            while re_spe.search(line) != None:
                start, stop = re_spe.search(line).span()
                spe_str = line[start:stop]
                if spe_str[1] == 'd':
                    base = 10
                elif spe_str[1] == 'h' or spe_str[1] == 'x':
                    base = 16
                elif spe_str[1] == 'o':
                    base = 8
                caract = chr(int(spe_str[2:], base))
                line = line[:start] + caract + line[stop:]
            if line != '':
                if line[-1] != '\x0C':
                    line += '\r'
                file_data[current_index].raw_data += line
    for data in file_data.data:
        if data.dType == 'program':
            data.use_base = is_base_program(data.raw_data)
    return file_data

def ctf_save(filename, file_data, export):
    """Save a file in the Calculator Text Format"""
    ctf_file = open(filename, 'w')
    ctn = ''
    # For each data, convert in the ctf format
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for data in [file_data.data[index] for index in index_list]:
        block_type = ''
        if data.dType == 'program':
            block_type = '@@Program'
        if block_type != '':
            # Set the header and metadata
            ctn += block_type + ' "' + ctfdic.ctf_encode(data.name, True) + '"'
            if data.password != None:
                ctn += ' (' + ctfdic.ctf_encode(data.password, True) + ')'
            ctn += '\n'
            # Write ctf data
            ctn += ctfdic.ctf_encode(data.raw_data) + 2 * '\n'
    # join in a newcat file
    ctf_file.write(ctn)
    ctf_file.close()

# Device manager

def receive(device_tool = 'default', port = 'default'):
    """Receive data from a device using device_tool, by port.

    device_tool and port default values depend of your OS
    Raise a ToolNotFound error if the device_tool is not in the
    path, raise a ToolNotConfigured if the device_tool 'casio_other'
    is not set, or if the default port dont work.
    device_tool allowed values are keys of formatType with value 'device'
    Return None if the file cannot be open
    Return a file_data if the transfert success
    
    """
    if device_tool == 'default':
        if os.name == 'posix':
            # Use cafix
            device_tool = 'casio_cafix'
        else :
            # Use CaS
            device_tool = 'casio_cas'
    if port == 'default':
        if os.name == 'posix':
            # Use /dev/ttyS0
            port = '/dev/ttyS0'
        else :
            # Use COM1
            port = 'com1'
    
    try :
        return openFunc[device_tool](port)
    except KeyError:
        raise ToolNotFound(device_tool)

def send(file_data, device_tool='default', port='default', export=False, 
         ignore_warnings=False):
    """Send data to a device

    If all data aren't managed by the device/transfer tool, raise
    DataNoManaged with the list
    of data type not managed (['all'] for all ;-) )
    Use ignore_warnings = True to use already this transfer tool.

    """
    if device_tool == 'default':
        if os.name == 'posix':
            # Use cafix
            device_tool = 'casio_cafix'
        else :
            # Use CaS
            device_tool = 'casio_cas'
    if port == 'default':
        if os.name == 'posix':
            # Use /dev/ttyS0
            port = '/dev/ttyS0'
        else :
            # Use COM1
            port = 'com1'
    if not ignore_warnings :
        no_managed_data = check_no_managed_data(file_data, device_tool, export)
        if no_managed_data != []:
            raise DataNoManaged(no_managed_data)
    try :
        saveFunc[device_tool](file_data, port, export)
    except KeyError:
        raise ToolNotFound(device_tool)

# For cafix

def cafix_receive(port = None):
    """Import data in file_data from a casio with cafix.

    Raise NoCommand if cafix isn't in the path
    Raise ToolNotConfigured if /dev/casio doesn't exist
    Return a file_data if the open is right
    
    """
    # Check cafix
    if len([path for path in os.defpath.split(os.pathsep) \
            if os.path.exists(os.path.join(path, 'cafix'))]) == 0:
        raise NoCommand('cafix')
    # Check /dev/casio
    if not os.path.exists('/dev/casio'):
        raise ToolNotConfigured
    # Choice a temp directory and make it
    directory = tempfile.mkdtemp()
    cafix_receive_interface(directory)
    # import all cafix files
    file_data = FileData()
    for file_name in [file_name for file_name in os.listdir(directory)\
                 if os.path.isfile(os.path.join(directory, file_name)) \
                    and file_name[-6:] == '.cafix']:
        (new_index,) = file_data.add(os.path.join(directory, file_name))
        file_data.data[new_index].date = datetime.date.today()
    return file_data

def cafix_send(file_data, port, export=False):
    """Transfert data from file_data to the casio using cafix.

    Raise NoCommand if cafix isn't found in the path
    Return ToolNotConfigured if /dev/casio doesn't exist
    Return None if all is Ok
    
    """
    # Check cafix
    if len([path for path in os.defpath.split(os.pathsep) \
            if os.path.exists(os.path.join(path, 'cafix'))]) == 0:
        raise NoCommand('cafix')
    # Check /dev/casio
    if not os.path.exists('/dev/casio'):
        raise ToolNotConfigured
    # Choice a temp directory and make it
    directory = tempfile.mkdtemp()
    # Export data in this folder
    if export:
        index_list = file_data.export
    else:
        index_list = range(len(file_data.data))
    for index in index_list:
        cafix_file = open(os.path.join(directory, file_data.data[index].name),
                         'w')
        cafix_file.write(file_data.data[index].raw_data)
        cafix_file.close()
    cafix_send_interface(directory)

def cafix_send_interface(tmp_dir):
    """Send data from tmpDir to the casio using cafix
    To be over writted for non-CLI interfaces
    
    """
    # Get the current dir
    cur_dir = os.path.abspath(os.path.curdir)
    print "Run your calculator, and set it in the state 'Receiving'."
    print "Next press Enter."
    raw_input()
    os.system("cd " + tmp_dir + " && cafix -s *")
    os.system("cd " + cur_dir)

def cafix_receive_interface(tmp_dir):
    """Receive data from a casio with cafix in tmpDir
    To be over writted for non-CLI interfaces 

    """
    cur_dir = os.path.abspath(os.path.curdir)
    print "You can send data from the calculator"
    os.system("cd " + tmp_dir + " && cafix -r")
    os.system("cd " + cur_dir)

# For Cas
def cas_receive(port):
    """Import data in file_data from a casio with CaS.

    Raise NoCommand if CaS isn't in the path
    Return a file_data if the open is right
    
    """
    # Check CaS
    if len([path for path in os.defpath.split(os.pathsep) \
            if os.path.exists(os.path.join(path, 'cas'))]) == 0:
        raise NoCommand('cas')
    # Choice a temp filename
    ctf_file = tempfile.mkstemp('.ctf')[1]
    cas_receive_interface(ctf_file, port)
    # import the ctf file
    return open_file(ctf_file, 'ctf')

def cas_receive_interface(ctf_file, port):
    """Receive data from a casio with CaS in ctf_file
    To be over writted for non-CLI interfaces 

    """
    print "You can send data from the calculator"
    os.system("cas -i " + port + " -o '" + ctf_file + "'")

def cas_send(file_data, port, export=False):
    """Transfert data from file_data to the casio using CaS.

    Raise NoCommand if cas isn't found in the path
    Return None if all is Ok
    
    """
    # Check CaS
    if len([path for path in os.defpath.split(os.pathsep) \
            if os.path.exists(os.path.join(path, 'cas'))]) == 0:
        raise NoCommand('cas')
    # Choice a temp ctf file and make it
    ctf_file = tempfile.mkstemp('.ctf')[1]
    file_data.save(ctf_file, 'ctf', ignore_warning = True, export = export)
    cas_send_interface(ctf_file, port)

def cas_send_interface(ctf_file, port):
    """Send data from ctf_file to the casio using CaS
    To be over writted for non-CLI interfaces
    
    """
    print "Run your calculator, and set it in the state 'Receiving'."
    print "Next press Enter."
    raw_input()
    os.system("cas -i '" + ctf_file + "' -o " + port)

# For another tool

def other_receive(port):
    """Import data in file_data from a casio with another tool.

    Return a file_data if the open is right
    Raise ToolNotConfigured if the tool is not set
    
    """
    # Check if the tool is set
    if _other_tool_receive == None:
        raise ToolNotConfigured
    # Choice a temp filename
    try:
        tmp_file = tempfile.mkstemp("." + formatExt[_other_tool_format][0])[1]
    except KeyError:
        raise FormatNotFound
    cmdline = _other_tool_receive.replace(r'%port', port).replace(r'%file', 
                                                                  tmp_file)
    other_receive_interface(cmdline)
    # import the temp file
    return open_file(tmp_file)

def other_receive_interface(cmdline):
    """Receive data from a casio with another tool
    To be over writted for non-CLI interfaces 

    """
    print "You can send data from the calculator"
    if os.system(cmdline):
        raise ToolNotConfigured
    #[term [-c]] program [options] %port %file

def other_send(file_data, port, export=False):
    """Transfert data from file_data to the casio using another transfer tool.

    Raise ToolNotConfigured if the tool is not set
    Return None if all is Ok
    
    """
    # Check if the tool is set
    if _other_tool_receive == None:
        raise ToolNotConfigured
    # Choice a temp filename and make it
    try:
        tmp_file = tempfile.mkstemp("." + formatExt[_other_tool_format][0])[1]
    except KeyError:
        raise FormatNotFound
    cmdline = _other_tool_receive.replace(r'%port', port).replace(r'%file', 
                                                                  tmp_file)
    file_data.save(tmp_file, _other_tool_format, ignore_warnings = True, 
                   export = export)
    other_send_interface(cmdline)

def other_send_interface(cmdline):
    """Send data to the casio using another tool
    To be over writted for non-CLI interfaces
    
    """
    print "Run your calculator, and set it in the state 'Receiving'."
    print "Next press Enter."
    raw_input()
    if os.system(cmdline):
        raise ToolNotConfigured

# Misc

def is_base_program(raw_data):
    """Return if the raw_data program use base functions
    This function is used to detect if the program may have the
    BASE flag. It must be used only if casetta is not able to detect
    the BASE flag in anoter way.
    
    """
    reg_exp = ''
    for spe_car in catdic.base_car:
        reg_exp += '[^\f7\7f]' + spe_car + '|'
    reg_exp = reg_exp[:-1]
    base_pattern = re.compile(reg_exp)
    if base_pattern.search(raw_data) == None:
        return False
    else:
        return True

def set_other_tool(format=None, receive_cmdline=None, send_cmdline=None):
    """Set settings of the other transfer tool
    
    format is the ID of a format read and write by both casetta and the
    transfert tool. receive_cmdline and send_cmdline can use %port and
    %file variable."""
    global _other_tool_format, _other_tool_receive, _other_tool_send
    _other_tool_format = format
    _other_tool_receive = receive_cmdline
    _other_tool_send = send_cmdline
    try: 
        dataManaged['casio_other'] = dataManaged[_other_tool_format]
    except KeyError:
        raise FormatNotFound
    openFunc['casio_other'] = other_receive
    saveFunc['casio_other'] = other_send

# Data type list :
dataType = ['unknown', 'none', 'program', 'backup']
# Name system of a data (regular expression pattern):
dataName = {'unknown':r'^.', 'none':r'.', \
            'program':r'^.{1,8}$', 'backup':r'^.'}

# list of formats
formats = ['unknown', 'cafix', 'cat', 'fxi', 'newcat', 'fxb', \
           'fxm', 'fxd', 'g1r', 'ctf', 'casio_cafix', 'casio_other']
# witch data is managed by a format
dataManaged = {'unknown':[], \
               'casio_cafix':['unknown', 'program', 'backup'], \
               'cafix':['unknown', 'program', 'backup'], \
               'cat': ['program'], \
               'fxi':['program'], \
               'newcat':['program', 'backup'], \
               'fxb':['backup'], \
               'fxm':['backup'], \
               'fxd':['program'], \
               'g1r':['program'], \
               'ctf':['program']}
# type of format ('file' or 'device')
formatType = {'unknown': 'none', 'casio_cafix': 'device', 'cafix': 'file', \
              'cat': 'file', 'fxi': 'file', 'newcat': 'file', \
              'fxb':'file', 'fxm':'file', 'fxd':'file', 'g1r':'file', \
              'ctf':'file', 'casio_other':'device'}
# what is the function to open a file_data in a format ?
# (arguments for file formats: filename ; return a file_data)
openFunc = {'casio_cafix': cafix_receive, 'cafix': cafix_open, \
            'cat': cat_open, 'fxi': fxi_open, 'newcat': newcat_open, \
            'fxb': fxb_open, 'fxm': fxm_open, 'fxd': fxd_open, \
            'g1r': g1r_open, 'ctf': ctf_open}
# what is the function to save a file_data in a format
# (arguments fo file formats: target_filename, file_data, export=False)
saveFunc = {'casio_cafix': cafix_send, 'cafix': cafix_save, 'cat': cat_save, \
            'newcat': newcat_save, 'fxb': fxb_save, 'ctf': ctf_save}
# list of the extension of a file format (without the dot)
formatExt = {'cafix':['cafix'], 'cat':['cat'], 'fxi':['fxi'], \
             'newcat':['newcat'], 'fxb':['fxb'], 'fxm':['fxm'], \
             'fxd':['fxd'], 'g1r':['g1r'], 'ctf':['ctf', 'txt']}

# vim:ai:et:sw=4:ts=4:sts=4:tw=78:fenc=utf-8
